@@ -20,6 +20,7 @@
 #define _PROCMAN_PROCMAN_H_
 
 
+#include <gtkmm.h>
 
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <glib.h>
@@ -29,6 +30,8 @@
 
 #include <time.h>
 
+#include <map>
+
 typedef struct _ProcConfig ProcConfig;
 struct ProcInfo;
 struct ProcData;
@@ -80,25 +83,37 @@ struct _ProcConfig
 	GdkColor	bg_color;
 	GdkColor	frame_color;
 	gint 		num_cpus;
+	bool solaris_mode;
 };
 
 
 class ProcInfo
 {
-	ProcInfo& operator=(const ProcInfo&);
-	ProcInfo(const ProcInfo&);
+	/* undefined */ ProcInfo& operator=(const ProcInfo&);
+	/* undefined */ ProcInfo(const ProcInfo&);
  public:
+
+	// TODO: use a set instead
+	// sorted by pid. The map has a nice property : it is sorted
+	// by pid so this helps a lot when looking for the parent node
+	// as ppid is nearly always < pid.
+	typedef std::map<pid_t, ProcInfo*> List;
+	typedef List::iterator Iterator;
+
+	static List all;
+
+	static ProcInfo* find(pid_t pid);
+	static Iterator begin() { return ProcInfo::all.begin(); }
+	static Iterator end() { return ProcInfo::all.end(); }
+
+
 	ProcInfo(pid_t pid);
 	~ProcInfo();
 	// adds one more ref to icon
-	void set_icon(GdkPixbuf *icon);
+	void set_icon(Glib::RefPtr<Gdk::Pixbuf> icon);
 
 	GtkTreeIter	node;
-	GtkTreePath	*path;
-	ProcInfo	*parent;
-	GSList		*children;
-
-	GdkPixbuf	*pixbuf;
+	Glib::RefPtr<Gdk::Pixbuf> pixbuf;
 	gchar		*tooltip;
 	gchar		*name;
 	gchar		*user; /* allocated with g_string_chunk, don't free it ! */
@@ -123,14 +138,12 @@ class ProcInfo
 	// wnck gives an unsigned long
 	unsigned long	memxserver;
 
-	guint		pid;
+	const guint	pid;
+	guint		ppid;
 	guint		uid;
 
 	guint8		pcpu; /* 0% - 100% */
 	gint8		nice;
-
-	guint		is_visible	: 1;
-	guint		is_running	: 1;
 };
 
 struct ProcData
@@ -158,32 +171,6 @@ struct ProcData
 	guint		timeout;
 	guint		disk_timeout;
 
-/*
-   'info' is GList, which has very slow lookup. On each update, glibtop
-   retrieves the full system pid list. To update the table,
-
-   foreach pid in glibtop pid list:
-	- if there's not ProcInfo with pid, add a new ProcInfo to 'info'
-	- else, update the current ProcInfo
-
-   which is very inefficient, because if a process is running at time t,
-   it's very likely to be running at t+1.
-   So if is length('info') = N, N lookups are performed on each update.
-   Average lookup iterates over N / 2 ProcInfo to find that a pid already
-   has a ProcInfo.
-   -> cost = (N x N) / 2
-
-   With 200 processes, an update costs about 20000 g_list_next(), which
-   is huge because we only want to know if we have this pid or not.
-
-   So 'pids' is a hastable to perform fast lookups.
-
-   TODO: 'info' and 'pids' could be merged
-*/
-
-	GList		*info;
-	GHashTable	*pids;
-
 	PrettyTable	pretty_table;
 
 	GConfClient	*client;
@@ -199,9 +186,9 @@ struct ProcData
 	SmoothRefresh  *smooth_refresh;
 };
 
-void		procman_save_config (ProcData *data) G_GNUC_INTERNAL;
-void		procman_save_tree_state (GConfClient *client, GtkWidget *tree, const gchar *prefix) G_GNUC_INTERNAL;
-gboolean	procman_get_tree_state (GConfClient *client, GtkWidget *tree, const gchar *prefix) G_GNUC_INTERNAL;
+void		procman_save_config (ProcData *data);
+void		procman_save_tree_state (GConfClient *client, GtkWidget *tree, const gchar *prefix);
+gboolean	procman_get_tree_state (GConfClient *client, GtkWidget *tree, const gchar *prefix);
 
 
 
