@@ -39,15 +39,18 @@
 #include <pwd.h>
 #include <time.h>
 
+#include <set>
+#include <list>
+
 #include <libgnomevfs/gnome-vfs-utils.h>
 
 #include "procman.h"
+#include "selection.h"
 #include "proctable.h"
 #include "callbacks.h"
 #include "prettytable.h"
 #include "util.h"
 #include "interface.h"
-#include "favorites.h"
 #include "selinux.h"
 extern "C" {
 #include "e_date.h"
@@ -57,6 +60,19 @@ static guint64 total_time = 1;
 static guint64 total_time_last = 1;
 
 
+
+ProcInfo::List ProcInfo::all;
+
+
+ProcInfo* ProcInfo::find(pid_t pid)
+{
+  Iterator it(ProcInfo::all.find(pid));
+  return (it == ProcInfo::all.end() ? NULL : it->second);
+}
+
+
+
+
 static gint
 sort_ints (GtkTreeModel *model, GtkTreeIter *itera, GtkTreeIter *iterb, gpointer data)
 {
@@ -70,21 +86,6 @@ sort_ints (GtkTreeModel *model, GtkTreeIter *itera, GtkTreeIter *iterb, gpointer
 	g_assert(infob);
 
 	switch (col) {
-	case COL_VMSIZE:
-		return PROCMAN_RCMP(infoa->vmsize, infob->vmsize);
-
-	case COL_MEMRES:
-		return PROCMAN_RCMP(infoa->memres, infob->memres);
-
-	case COL_MEMWRITABLE:
-		return PROCMAN_RCMP(infoa->memwritable, infob->memwritable);
-
-	case COL_MEMSHARED:
-		return PROCMAN_RCMP(infoa->memshared, infob->memshared);
-
-	case COL_MEMXSERVER:
-		return PROCMAN_RCMP(infoa->memxserver, infob->memxserver);
-
 	case COL_CPU_TIME:
 		return PROCMAN_RCMP(infoa->cpu_time_last, infob->cpu_time_last);
 
@@ -93,10 +94,6 @@ sort_ints (GtkTreeModel *model, GtkTreeIter *itera, GtkTreeIter *iterb, gpointer
 
 	case COL_CPU:
 		return PROCMAN_RCMP(infoa->pcpu, infob->pcpu);
-
-	case COL_MEM:
-		return PROCMAN_RCMP(infoa->mem, infob->mem);
-
 	default:
 		g_assert_not_reached();
 		return 0;
@@ -304,11 +301,11 @@ proctable_new (ProcData * const procdata)
 				    G_TYPE_STRING,	/* Process Name */
 				    G_TYPE_STRING,	/* User		*/
 				    G_TYPE_STRING,	/* Status	*/
-				    G_TYPE_STRING,	/* VM Size	*/
-				    G_TYPE_STRING,	/* Resident Memory */
-				    G_TYPE_STRING,	/* Writable Memory */
-				    G_TYPE_STRING,	/* Shared Memory */
-				    G_TYPE_STRING,	/* X Server Memory */
+				    G_TYPE_ULONG,	/* VM Size	*/
+				    G_TYPE_ULONG,	/* Resident Memory */
+				    G_TYPE_ULONG,	/* Writable Memory */
+				    G_TYPE_ULONG,	/* Shared Memory */
+				    G_TYPE_ULONG,	/* X Server Memory */
 				    G_TYPE_UINT,	/* % CPU	*/
 				    G_TYPE_STRING,	/* CPU time	*/
 				    G_TYPE_STRING,	/* Started	*/
@@ -316,7 +313,7 @@ proctable_new (ProcData * const procdata)
 				    G_TYPE_UINT,	/* ID		*/
 				    G_TYPE_STRING,	/* Security Context */
 				    G_TYPE_STRING,	/* Arguments	*/
-				    G_TYPE_STRING,	/* Memory       */
+				    G_TYPE_ULONG,	/* Memory       */
 				    GDK_TYPE_PIXBUF,	/* Icon		*/
 				    G_TYPE_POINTER,	/* ProcInfo	*/
 				    G_TYPE_STRING	/* Sexy tooltip */
@@ -356,8 +353,39 @@ proctable_new (ProcData * const procdata)
 
 
 	for (i = COL_USER; i <= COL_MEM; i++) {
-		cell_renderer = gtk_cell_renderer_text_new ();
 
+		GtkCellRenderer *cell;
+		GtkTreeViewColumn *col;
+
+		cell = gtk_cell_renderer_text_new();
+		col = gtk_tree_view_column_new();
+		gtk_tree_view_column_pack_start(col, cell, TRUE);
+		gtk_tree_view_column_set_title(col, _(titles[i]));
+		gtk_tree_view_column_set_resizable(col, TRUE);
+		gtk_tree_view_column_set_sort_column_id(col, i);
+		gtk_tree_view_column_set_reorderable(col, TRUE);
+		gtk_tree_view_append_column(GTK_TREE_VIEW(proctree), col);
+
+		// type
+		switch (i) {
+		case COL_VMSIZE:
+		case COL_MEMRES:
+		case COL_MEMWRITABLE:
+		case COL_MEMSHARED:
+		case COL_MEMXSERVER:
+		case COL_MEM:
+		  gtk_tree_view_column_set_cell_data_func(col, cell,
+							  &procman::size_cell_data_func,
+							  GUINT_TO_POINTER(i),
+							  NULL);
+		  break;
+
+		default:
+		  gtk_tree_view_column_set_attributes(col, cell, "text", i, NULL);
+		  break;
+		}
+
+		// xaling
 		switch(i)
 		{
 		case COL_VMSIZE:
@@ -370,28 +398,20 @@ proctable_new (ProcData * const procdata)
 		case COL_PID:
 		case COL_CPU_TIME:
 		case COL_MEM:
-			g_object_set(G_OBJECT(cell_renderer),
-				     "xalign", 1.0f,
-				     NULL);
-		}
-
-		column = gtk_tree_view_column_new ();
-		gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
-		gtk_tree_view_column_set_attributes (column, cell_renderer,
-						     "text", i,
-						     NULL);
-		gtk_tree_view_column_set_title (column, _(titles[i]));
-		gtk_tree_view_column_set_sort_column_id (column, i);
-		gtk_tree_view_column_set_resizable (column, TRUE);
-
-		switch (i) {
-		case COL_ARGS:
-			gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
+			g_object_set(G_OBJECT(cell), "xalign", 1.0f, NULL);
 			break;
 		}
 
-		gtk_tree_view_column_set_min_width (column, 1);
-		gtk_tree_view_append_column (GTK_TREE_VIEW (proctree), column);
+		// sizing
+		switch (i) {
+		case COL_ARGS:
+		  gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_FIXED);
+		  gtk_tree_view_column_set_min_width(col, 150);
+		  break;
+		default:
+		  gtk_tree_view_column_set_min_width(column, 20);
+		  break;
+		}
 	}
 
 	gtk_container_add (GTK_CONTAINER (scrolled), proctree);
@@ -401,15 +421,9 @@ proctable_new (ProcData * const procdata)
 	{
 		switch(i)
 		{
-		case COL_VMSIZE:
-		case COL_MEMRES:
-		case COL_MEMWRITABLE:
-		case COL_MEMSHARED:
-		case COL_MEMXSERVER:
 		case COL_CPU_TIME:
 		case COL_START_TIME:
 		case COL_CPU:
-		case COL_MEM:
 			gtk_tree_sortable_set_sort_func (
 				GTK_TREE_SORTABLE (model),
 				i,
@@ -447,21 +461,12 @@ proctable_new (ProcData * const procdata)
 }
 
 
-static void
-proctable_free_info(ProcInfo *info)
+ProcInfo::~ProcInfo()
 {
-	g_assert(info != NULL);
-
-	if(info->pixbuf) {
-		g_object_unref (info->pixbuf);
-	}
-
-	g_free (info->name);
-	g_free(info->tooltip);
-	g_free (info->arguments);
-	g_free (info->security_context);
-	g_slist_free (info->children);
-	g_slice_free(ProcInfo, info);
+  g_free(this->name);
+  g_free(this->tooltip);
+  g_free(this->arguments);
+  g_free(this->security_context);
 }
 
 
@@ -473,27 +478,22 @@ get_process_status (ProcInfo *info, const glibtop_proc_state *buf)
 	{
 	case GLIBTOP_PROCESS_RUNNING:
 		info->status = _("Running");
-		info->is_running = TRUE;
 		break;
 
 	case GLIBTOP_PROCESS_STOPPED:
 		info->status = _("Stopped");
-		info->is_running = FALSE;
 		break;
 
 	case GLIBTOP_PROCESS_ZOMBIE:
 		info->status = _("Zombie");
-		info->is_running = FALSE;
 		break;
 
 	case GLIBTOP_PROCESS_UNINTERRUPTIBLE:
 		info->status = _("Uninterruptible");
-		info->is_running = FALSE;
 		break;
 
 	default:
 		info->status = _("Sleeping");
-		info->is_running = FALSE;
 		break;
 	}
 }
@@ -596,24 +596,6 @@ get_process_memory_info(ProcInfo *info)
 
 
 
-ProcInfo *
-proctable_find_process (guint pid, ProcData *procdata)
-{
-	return static_cast<ProcInfo*>(g_hash_table_lookup(procdata->pids, GINT_TO_POINTER(pid)));
-}
-
-
-static ProcInfo *
-find_parent (ProcData *procdata, guint pid)
-{
-	/* ignore init as a parent process */
-	if (pid <= 1)
-		return NULL;
-
-	return proctable_find_process(pid, procdata);
-}
-
-
 static inline unsigned divide(unsigned *q, unsigned *r, unsigned d)
 {
 	*q = *r / d;
@@ -654,82 +636,67 @@ format_duration_for_display(unsigned centiseconds)
 
 
 static void
-update_info_mutable_cols(GtkTreeStore *store, ProcData *procdata, ProcInfo *info)
+update_info_mutable_cols(ProcInfo *info)
 {
-	gchar *vmsize, *memres, *memwritable, *memshared, *memxserver,
-		*cpu_time, *start_time, *mem;
+	ProcData * const procdata = ProcData::get_instance();
+	GtkTreeModel *model;
+	model = gtk_tree_view_get_model(GTK_TREE_VIEW(procdata->tree));
 
-	vmsize	   = SI_gnome_vfs_format_file_size_for_display (info->vmsize);
-	memres	   = SI_gnome_vfs_format_file_size_for_display (info->memres);
-	memwritable = SI_gnome_vfs_format_file_size_for_display (info->memwritable);
-	memshared  = SI_gnome_vfs_format_file_size_for_display (info->memshared);
-	memxserver = SI_gnome_vfs_format_file_size_for_display (info->memxserver);
-	mem = SI_gnome_vfs_format_file_size_for_display(info->mem);
+	gchar *cpu_time, *start_time;
 
+	// expects centiseconds
 	cpu_time = format_duration_for_display(100 * info->cpu_time_last / procdata->frequency);
 
 	/* FIXME: does it worths it to display relative to $now date ?
 	   absolute date wouldn't required to be updated on every refresh */
 	start_time = procman_format_date_for_display(info->start_time);
 
-	gtk_tree_store_set (store, &info->node,
+	gtk_tree_store_set(GTK_TREE_STORE(model), &info->node,
 			    COL_STATUS, info->status,
 			    COL_USER, info->user,
-			    COL_VMSIZE, vmsize,
-			    COL_MEMRES, memres,
-			    COL_MEMWRITABLE, memwritable,
-			    COL_MEMSHARED, memshared,
-			    COL_MEMXSERVER, memxserver,
+			    COL_VMSIZE, info->vmsize,
+			    COL_MEMRES, info->memres,
+			    COL_MEMWRITABLE, info->memwritable,
+			    COL_MEMSHARED, info->memshared,
+			    COL_MEMXSERVER, info->memxserver,
 			    COL_CPU, guint(info->pcpu),
 			    COL_CPU_TIME, cpu_time,
 			    COL_START_TIME, start_time,
 			    COL_NICE, gint(info->nice),
-			    COL_MEM, mem,
+			    COL_MEM, info->mem,
 			    -1);
 
 	/* FIXME: We don't bother updating COL_SECURITYCONTEXT as it can never change.
 	   even on fork ? */
-	g_free (vmsize);
-	g_free (memres);
-	g_free (memwritable);
-	g_free (memshared);
-	g_free (memxserver);
 	g_free (cpu_time);
 	g_free(start_time);
-	g_free(mem);
 }
 
 
 
-void
-insert_info_to_tree (ProcInfo *info, ProcData *procdata)
+static void
+insert_info_to_tree (ProcInfo *info, ProcData *procdata, bool forced = false)
 {
 	GtkTreeModel *model;
 
-	/* Don't show process if it is not running */
-	if ((procdata->config.whose_process == ACTIVE_PROCESSES) &&
-	    (!info->is_running))
-		return;
-
-	/* Don't show processes that user has blacklisted */
-	if (is_process_blacklisted (procdata, info->name))
-	{
-		info->is_blacklisted = TRUE;
-		return;
-	}
-	info->is_blacklisted = FALSE;
-
 	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-	if (info->parent && procdata->config.show_tree && info->parent->is_visible) {
-		GtkTreePath *parent_node = gtk_tree_model_get_path (model, &info->parent->node);
 
-		gtk_tree_store_insert (GTK_TREE_STORE (model), &info->node, &info->parent->node, 0);
-		if (!gtk_tree_view_row_expanded (GTK_TREE_VIEW (procdata->tree), parent_node))
-			gtk_tree_view_expand_row (GTK_TREE_VIEW (procdata->tree),
-						  parent_node,
-						  FALSE);
+	if (procdata->config.show_tree) {
 
-		gtk_tree_path_free (parent_node);
+	  ProcInfo *parent = 0;
+
+	  if (not forced)
+	    parent = ProcInfo::find(info->ppid);
+
+	  if (parent) {
+	    GtkTreePath *parent_node = gtk_tree_model_get_path(model, &parent->node);
+	    gtk_tree_store_insert(GTK_TREE_STORE(model), &info->node, &parent->node, 0);
+
+	    if (!gtk_tree_view_row_expanded(GTK_TREE_VIEW(procdata->tree), parent_node))
+	      gtk_tree_view_expand_row(GTK_TREE_VIEW(procdata->tree), parent_node, FALSE);
+	    gtk_tree_path_free(parent_node);
+	  } else
+	    gtk_tree_store_insert(GTK_TREE_STORE(model), &info->node, NULL, 0);
 	}
 	else
 		gtk_tree_store_insert (GTK_TREE_STORE (model), &info->node, NULL, 0);
@@ -741,7 +708,7 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 
 	gtk_tree_store_set (GTK_TREE_STORE (model), &info->node,
 			    COL_POINTER, info,
-			    COL_PIXBUF, info->pixbuf,
+			    COL_PIXBUF, info->pixbuf->gobj(),
 			    COL_NAME, info->name,
 			    COL_ARGS, info->arguments,
 			    COL_TOOLTIP, info->tooltip,
@@ -749,280 +716,270 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 			    COL_SECURITYCONTEXT, info->security_context,
 			    -1);
 
-	update_info_mutable_cols(GTK_TREE_STORE (model), procdata, info);
-
-	info->is_visible = TRUE;
+	procman_debug("inserted %d%s", info->pid, (forced ? " (forced)" : ""));
 }
 
 
 /* Removing a node with children - make sure the children are queued
 ** to be readded.
 */
+template<typename List>
 static void
-remove_children_from_tree (ProcData *procdata, GtkTreeModel *model,
-			   GtkTreeIter *parent, GPtrArray *addition_list)
+remove_info_from_tree (ProcData *procdata, GtkTreeModel *model,
+		       ProcInfo *current, List &orphans, unsigned lvl = 0)
 {
-	do {
-		ProcInfo *child_info;
-		GtkTreeIter child;
+  GtkTreeIter child_node;
 
-		if (gtk_tree_model_iter_children (model, &child, parent))
-			remove_children_from_tree (procdata, model, &child, addition_list);
+  if (std::find(orphans.begin(), orphans.end(), current) != orphans.end()) {
+    procman_debug("[%u] %d already removed from tree", lvl, int(current->pid));
+    return;
+  }
 
-		gtk_tree_model_get (model, parent, COL_POINTER, &child_info, -1);
-		if (child_info) {
-			if (procdata->selected_process == child_info)
-				procdata->selected_process = NULL;
-			g_ptr_array_add(addition_list, child_info);
-			child_info->is_visible = FALSE;
-		}
-	} while (gtk_tree_model_iter_next (model, parent));
+  procman_debug("[%u] pid %d, %d children", lvl, int(current->pid),
+		gtk_tree_model_iter_n_children(model, &current->node));
+
+  // it is not possible to iterate&erase over a treeview so instead we
+  // just pop one child after another and recursively remove it and
+  // its children
+
+  while (gtk_tree_model_iter_children(model, &child_node, &current->node)) {
+    ProcInfo *child = 0;
+    gtk_tree_model_get(model, &child_node, COL_POINTER, &child, -1);
+    remove_info_from_tree(procdata, model, child, orphans, lvl + 1);
+  }
+
+  g_assert(not gtk_tree_model_iter_has_child(model, &current->node));
+
+  if (procdata->selected_process == current)
+    procdata->selected_process = NULL;
+
+  orphans.push_back(current);
+  gtk_tree_store_remove(GTK_TREE_STORE(model), &current->node);
+  procman::poison(current->node, 0x69);
 }
 
 
-void
-remove_info_from_tree (ProcInfo *info, ProcData *procdata)
-{
-	GtkTreeModel *model;
-
-	g_assert(info);
-
-	if (!info->is_visible)
-		return;
-
-	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-
-	if (procdata->selected_process == info)
-		procdata->selected_process = NULL;
-
-	gtk_tree_store_remove (GTK_TREE_STORE (model), &info->node);
-
-	info->is_visible = FALSE;
-}
-
-
-static void
-remove_info_from_list (ProcInfo *info, ProcData *procdata)
-{
-	GSList *child;
-	ProcInfo * const parent = info->parent;
-
-	/* Remove info from parent list */
-	if (parent)
-		parent->children = g_slist_remove (parent->children, info);
-
-	/* Add any children to parent list */
-	for(child = info->children; child; child = g_slist_next (child)) {
-		ProcInfo *child_info = static_cast<ProcInfo*>(child->data);
-		child_info->parent = parent;
-	}
-
-	if(parent) {
-		parent->children = g_slist_concat(parent->children,
-						  info->children);
-		info->children = NULL;
-	}
-
-	/* Remove from list */
-	procdata->info = g_list_remove (procdata->info, info);
-	g_hash_table_remove(procdata->pids, GINT_TO_POINTER(info->pid));
-	proctable_free_info(info);
-}
-
 
 static void
 update_info (ProcData *procdata, ProcInfo *info)
 {
 	glibtop_proc_state procstate;
+	glibtop_proc_uid procuid;
+	glibtop_proc_time proctime;
 
 	glibtop_get_proc_state (&procstate, info->pid);
+	glibtop_get_proc_uid (&procuid, info->pid);
+	glibtop_get_proc_time (&proctime, info->pid);
+
 	get_process_status (info, &procstate);
 
-	if (procdata->config.whose_process == ACTIVE_PROCESSES)	{
+	get_process_memory_info(info);
 
-		/* process started running */
-		if (info->is_running && (!info->is_visible)) {
-			insert_info_to_tree (info, procdata);
-			return;
-		}
-		/* process was running but not anymore */
-		else if ((!info->is_running) && info->is_visible) {
-			remove_info_from_tree (info, procdata);
-			return;
-		}
-		else if (!info->is_running)
-			return;
-	}
+	get_process_user(procdata, info, procstate.uid);
 
-	if (info->is_visible) {
-		GtkTreeModel *model;
-		glibtop_proc_uid procuid;
-		glibtop_proc_time proctime;
+	info->pcpu = (proctime.rtime - info->cpu_time_last) * 100 / total_time;
+	info->pcpu = MIN(info->pcpu, 100);
 
-		glibtop_get_proc_uid (&procuid, info->pid);
-		glibtop_get_proc_time (&proctime, info->pid);
+	if (procdata->config.solaris_mode)
+	  info->pcpu /= procdata->config.num_cpus;
 
-		get_process_memory_info(info);
-		get_process_user(procdata, info, procstate.uid);
-
-		info->pcpu = (proctime.rtime - info->cpu_time_last) * 100 / total_time;
-		info->pcpu = MIN(info->pcpu, 100);
-
-		info->cpu_time_last = proctime.rtime;
-		info->nice = procuid.nice;
-
-
-		model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-
-		update_info_mutable_cols(GTK_TREE_STORE (model), procdata, info);
-	}
+	info->cpu_time_last = proctime.rtime;
+	info->nice = procuid.nice;
+	info->ppid = procuid.ppid;
 }
 
 
-static ProcInfo *
-procinfo_new (ProcData *procdata, gint pid)
+ProcInfo::ProcInfo(pid_t pid)
+  : tooltip(NULL),
+    name(NULL),
+    user(NULL),
+    arguments(NULL),
+    status(NULL),
+    security_context(NULL),
+    pid(pid),
+    uid(-1)
 {
-	ProcInfo *info;
+	ProcInfo * const info = this;
 	glibtop_proc_state procstate;
 	glibtop_proc_time proctime;
-	glibtop_proc_uid procuid;
 	glibtop_proc_args procargs;
 	gchar** arguments;
 
-	info = g_slice_new0(ProcInfo);
-
-	info->pid = pid;
-	info->uid = -1;
-
 	glibtop_get_proc_state (&procstate, pid);
-	glibtop_get_proc_uid (&procuid, pid);
 	glibtop_get_proc_time (&proctime, pid);
 	arguments = glibtop_get_proc_argv (&procargs, pid, 0);
 
 	/* FIXME : wrong. name and arguments may change with exec* */
 	get_process_name (info, procstate.cmd, arguments[0]);
-	get_process_user (procdata, info, procstate.uid);
 
 	info->tooltip = g_strjoinv(" ", arguments);
 	info->arguments = g_strescape(info->tooltip, "\\\"");
 	g_strfreev(arguments);
 
-	info->pcpu = 0;
 	info->cpu_time_last = proctime.rtime;
 	info->start_time = proctime.start_time;
-	info->nice = procuid.nice;
 
-	get_process_memory_info(info);
-	get_process_status (info, &procstate);
 	get_process_selinux_context (info);
 
-	info->pixbuf = procdata->pretty_table.get_icon(info->name, pid);
-
-	info->parent = find_parent (procdata, procuid.ppid);
-	if (info->parent) {
-		info->parent->children = g_slist_prepend (info->parent->children, info);
-	}
-
-	info->is_visible = FALSE;
-
-	return info;
+	info->pixbuf = ProcData::get_instance()->pretty_table.get_icon(info->name, pid);
 }
 
 
 
-static void cb_exclude(ProcInfo* info, GPtrArray *addition)
-{
-	g_ptr_array_remove_fast (addition, info);
-}
-
 
 static void
-refresh_list (ProcData *procdata, const unsigned *pid_list, const guint n)
+refresh_list (ProcData *procdata, const pid_t* pid_list, const guint n)
 {
-	GHashTable *pid_hash;
-	GList *list;
-	GPtrArray *addition_list = g_ptr_array_new ();
-	GPtrArray *removal_list = g_ptr_array_new ();
+  typedef std::list<ProcInfo*> ProcList;
+  ProcList addition;
+
 	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
 	guint i;
 
-	/* Add or update processes */
+	// Add or update processes in the process list
 	for(i = 0; i < n; ++i) {
-		ProcInfo *info;
+		ProcInfo *info = ProcInfo::find(pid_list[i]);
 
-		info = proctable_find_process (pid_list[i], procdata);
 		if (!info) {
-			info = procinfo_new (procdata, pid_list[i]);
-			g_ptr_array_add(addition_list, info);
-			procdata->info = g_list_prepend (procdata->info, info);
-			g_hash_table_insert(procdata->pids, GINT_TO_POINTER(info->pid), info);
-		}
-		else {
-			update_info (procdata, info);
+			info = new ProcInfo(pid_list[i]);
+			ProcInfo::all[info->pid] = info;
+			addition.push_back(info);
 		}
+
+		update_info (procdata, info);
 	}
 
 
-	/* Remove processes */
+	// Remove dead processes from the process list and from the
+	// tree. children are queued to be readded at the right place
+	// in the tree.
 
-	/* use a hash for fast lookup
-	   !NULL as data for every key */
+	const std::set<pid_t> pids(pid_list, pid_list + n);
 
-	pid_hash = g_hash_table_new(NULL, NULL);
-	for(i = 0; i < n; ++i)
-		g_hash_table_insert(pid_hash,
-				    GINT_TO_POINTER(pid_list[i]),
-				    GINT_TO_POINTER(1 /* !NULL */));
+	ProcInfo::Iterator it(ProcInfo::begin());
 
+	while (it != ProcInfo::end()) {
+	  ProcInfo * const info = it->second;
+	  ProcInfo::Iterator next(it);
+	  ++next;
 
-	for(list = procdata->info; list; list = g_list_next (list)) {
-		ProcInfo * const info = static_cast<ProcInfo*>(list->data);
+	  if (pids.find(info->pid) == pids.end()) {
+	    procman_debug("ripping %d", info->pid);
+	    remove_info_from_tree(procdata, model, info, addition);
+	    addition.remove(info);
+	    ProcInfo::all.erase(it);
+	    delete info;
+	  }
 
-		if(!g_hash_table_lookup(pid_hash, GINT_TO_POINTER(info->pid)))
-		{
-			g_ptr_array_add (removal_list, info);
-			/* remove all children from tree */
-			if (info->is_visible) {
-				GtkTreeIter child;
-				if (gtk_tree_model_iter_children (model, &child, &info->node))
-					remove_children_from_tree (procdata, model, &child, addition_list);
-			}
-		}
+	  it = next;
 	}
 
-	g_hash_table_destroy(pid_hash);
+	// INVARIANT
+	// pid_list == ProcInfo::all + addition
 
-	g_ptr_array_foreach(removal_list, (GFunc) cb_exclude, addition_list);
 
-	/* Add or remove processes from the tree */
-	g_ptr_array_foreach(removal_list,  (GFunc) remove_info_from_tree, procdata);
-	g_ptr_array_foreach(addition_list, (GFunc) insert_info_to_tree  , procdata);
+	if (procdata->config.show_tree) {
 
-	/* Remove processes from the internal GList */
-	g_ptr_array_foreach(removal_list,  (GFunc) remove_info_from_list, procdata);
+	// insert process in the tree. walk through the addition list
+	// (new process + process that have a new parent). This loop
+	// handles the dependencies because we cannot insert a process
+	// until its parent is in the tree.
 
-	g_ptr_array_free (addition_list, TRUE);
-	g_ptr_array_free (removal_list, TRUE);
+	std::set<pid_t> in_tree(pids);
+
+	for (ProcList::iterator it(addition.begin()); it != addition.end(); ++it)
+	  in_tree.erase((*it)->pid);
+
+
+	while (not addition.empty()) {
+	  procman_debug("looking for %d parents", int(addition.size()));
+	  ProcList::iterator it(addition.begin());
+
+	  while (it != addition.end()) {
+	    procman_debug("looking for %d's parent with ppid %d",
+			  int((*it)->pid), int((*it)->ppid));
+
+
+	    // inserts the process in the treeview if :
+	    // - it is init
+	    // - its parent is already in tree
+	    // - its parent is unreachable
+	    //
+	    // rounds == 2 means that addition contains processes with
+	    // unreachable parents
+	    //
+	    // FIXME: this is broken if the unreachable parent becomes active
+	    // i.e. it gets active or changes ower
+	    // so we just clear the tree on __each__ update
+	    // see proctable_update_list (ProcData * const procdata)
+
+
+	    if ((*it)->ppid == 0 or in_tree.find((*it)->ppid) != in_tree.end()) {
+	      insert_info_to_tree(*it, procdata);
+	      in_tree.insert((*it)->pid);
+	      it = addition.erase(it);
+	      continue;
+	    }
+
+	    ProcInfo *parent = ProcInfo::find((*it)->ppid);
+	    // if the parent is unreachable
+	    if (not parent) {
+		// or std::find(addition.begin(), addition.end(), parent) == addition.end()) {
+		insert_info_to_tree(*it, procdata, true);
+		in_tree.insert((*it)->pid);
+		it = addition.erase(it);
+		continue;
+	    }
+
+	    ++it;
+	  }
+	}
+	}
+	else { 
+	  // don't care of the tree
+	  for (ProcList::iterator it(addition.begin()); it != addition.end(); ++it)
+	    insert_info_to_tree(*it, procdata);
+	}
+
+
+	for (ProcInfo::Iterator it(ProcInfo::begin()); it != ProcInfo::end(); ++it)
+		update_info_mutable_cols(it->second);
 }
 
 
 void
 proctable_update_list (ProcData * const procdata)
 {
-	unsigned *pid_list;
+	pid_t* pid_list;
 	glibtop_proclist proclist;
 	glibtop_cpu cpu;
 	gint which, arg;
+	procman::SelectionMemento selection;
 
 	switch (procdata->config.whose_process) {
 	case ALL_PROCESSES:
-	case ACTIVE_PROCESSES:
 		which = GLIBTOP_KERN_PROC_ALL;
 		arg = 0;
 		break;
+
+	case ACTIVE_PROCESSES:
+		which = GLIBTOP_KERN_PROC_ALL | GLIBTOP_EXCLUDE_IDLE;
+		arg = 0;
+		if (procdata->config.show_tree)
+		  {
+		    selection.save(procdata->tree);
+		    proctable_clear_tree(procdata);
+		  }
+		break;
+
 	default:
 		which = GLIBTOP_KERN_PROC_UID;
 		arg = getuid ();
+		if (procdata->config.show_tree)
+		  {
+		    selection.save(procdata->tree);
+		    proctable_clear_tree(procdata);
+		  }
 		break;
 	}
 
@@ -1036,6 +993,8 @@ proctable_update_list (ProcData * const procdata)
 
 	refresh_list (procdata, pid_list, proclist.number);
 
+	selection.restore(procdata->tree);
+
 	g_free (pid_list);
 
 	/* proclist.number == g_list_length(procdata->info) == g_hash_table_size(procdata->pids) */
@@ -1073,20 +1032,10 @@ proctable_clear_tree (ProcData * const procdata)
 void
 proctable_free_table (ProcData * const procdata)
 {
-	GList *list = procdata->info;
+  for (ProcInfo::Iterator it(ProcInfo::begin()); it != ProcInfo::end(); ++it)
+    delete it->second;
 
-	while (list)
-	{
-		ProcInfo *info = static_cast<ProcInfo*>(list->data);
-		proctable_free_info(info);
-		list = g_list_next (list);
-	}
-
-	g_list_free (procdata->info);
-	procdata->info = NULL;
-
-	g_hash_table_destroy(procdata->pids);
-	procdata->pids = g_hash_table_new(NULL, NULL);
+  ProcInfo::all.clear();
 }
 
 
@@ -1109,19 +1058,13 @@ make_loadavg_string(void)
 
 
 void
-_ProcInfo::set_icon(GdkPixbuf* icon)
+ProcInfo::set_icon(Glib::RefPtr<Gdk::Pixbuf> icon)
 {
-  if (this->pixbuf)
-    g_object_unref(this->pixbuf);
-
   this->pixbuf = icon;
-  g_object_ref(icon);
 
-  if (this->is_visible) {
-    GtkTreeModel *model;
-    model = gtk_tree_view_get_model(GTK_TREE_VIEW(ProcData::get_instance()->tree));
-    gtk_tree_store_set(GTK_TREE_STORE(model), &this->node,
-		       COL_PIXBUF, this->pixbuf,
-		       -1);
-  }
+  GtkTreeModel *model;
+  model = gtk_tree_view_get_model(GTK_TREE_VIEW(ProcData::get_instance()->tree));
+  gtk_tree_store_set(GTK_TREE_STORE(model), &this->node,
+		     COL_PIXBUF, this->pixbuf->gobj(),
+		     -1);
 }
